/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   APACHE
* @PACKAGE  GEN_PROXY
*
* @FILE     gen_proxy.c
*
* @ABSTRACT
* This module is executed by the process httpd if the
* directive CCNProxyRemote is set to On in the file httpd.conf 
*
* This file contains:
*
* @INCLUDE

#include <stdio.h>
#include <sys/time.h>
#include <regex.h>
#include <string.h>
#include <stdlib.h>


#include "gcn_upcon.h"
#include "gcn_uptyp.h"
#include "gcn_upmac.h"
#include "cdx_upcon.h"
#include "cdx_uptyp.h"
#include "cdx_upext.h"
#include "osx_upext.h"
#include "evr_upext.h"

#ifdef CDX_GATEWAY
#include "cda_upcon.h"
#include "cda_upext.h"
#endif /* CDX_GATEWAY */

#include "osx_upcon.h"

#include "loglo_uptyp.h"
#include "gen_proxy_upcon.h"
#include "gen_proxy_prtyp.h"
#include "gen_proxy_prext.h"
#include "gen_proxy_prmes.h"


#define NB_PATTERN_PROXY 8

PROXYREMOTE proxy_remote_build[] = {
  {PORT_NUMBER_PATTERN_1,buildProxyPattern1},
  {PORT_NUMBER_PATTERN_2,buildProxyPattern2},
  {PORT_NUMBER_PATTERN_3,buildProxyPattern3},
  {PORT_NUMBER_PATTERN_4,buildProxyPattern4},
  {PORT_NUMBER_PATTERN_5,buildProxyPattern5},
  {PORT_NUMBER_PATTERN_6,buildProxyPattern6},
  {PORT_NUMBER_PATTERN_7,buildProxyPattern7},
  {PORT_NUMBER_PATTERN_8,buildProxyPattern8},
  {NULL,NULL}
};

static void usage ()
{
   printf ("usage:\n");
   printf ("   ccn_gen_proxyremote\n");
   printf ("   ccn_gen_proxyremote -v for verbose\n");
}

/*
* @FUNCTION validate_pattern
*
* @DESCRIPTION This function validate a pattern.
*
* @ABSTRACT
*
* @RETURNVALUE CSILONG
* validate_pattern returns 0 upon success
*
* @PARAMETER
* IN    CSICHAR  *szPattern        the regular expression.
* IN    CSICHAR  *szString         the string to evaluated.
* IN    CSILONG  szHtpps         the string to evaluated.
*
*/

static CSILONG validate_pattern(
  IN PROXYREMOTE *szProxyRemote,
  IN CSICHAR  *szString,
  IN PROXYHEADER  *szProxyInfo
)
{
  int lStatus;
  regex_t preg;
  regmatch_t pmatch[MAX_SUB_STRING];
  
  lStatus = regcomp(&preg,
                    szProxyRemote->szPattern,
                    REG_EXTENDED);

  if (lStatus != 0)
  {
    printf("ERROR: Unable to compile the pattern %s\n",szProxyRemote->szPattern);
  }
  else
  {
    lStatus = regexec(&preg,
                      szString,
                      MAX_SUB_STRING,
                      pmatch,
                      0);
    if (!lStatus)
    {
      szProxyRemote->szFnct(szString,preg.re_nsub,pmatch,szProxyInfo);      
    }
    regfree(&preg);
  }
  return lStatus;
}

/*
* @FUNCTION gen_proxy_remote
*
* @DESCRIPTION This generate on the STDOUT the proxy remote directive. 
*
* @ABSTRACT
*
* @RETURNVALUE CSILONG
* gen_proxy_remote returns 0 upon success
*
*
*/

static CSILONG gen_proxy_remote(
  CSILONG verbose
)
{
  CSILONG reasonCode;
  CSILONG returnCode = CSISUCCESS;
  CDXRESULT siteList = {CDXRESULT_DEFAULT};
  CSICHAR *szSiteName;
  CSIBOOLEAN boundOS=CSIFALSE;
  CSIBOOLEAN boundCDX=CSIFALSE;
  CSICHAR48 ccnWebDomain;
  CSICHAR48 envSiteName;
  CSICHAR48 envGtwName;
  CSICHAR256 strError;
  CDXRESULT sLocalGtwMode = {CDXRESULT_DEFAULT};

  if (NULL == getenv("CCN_SITE_NAME"))
  {
    printf("ERROR: Environment variable CCN_SITE_NAME not set\n");
    return 1;
  }
  memset(envSiteName, '\0', sizeof(envSiteName));
  GCN_STRCPY(envSiteName, getenv("CCN_SITE_NAME"));

  if (NULL == getenv("CCN_GATEWAY_NAME"))
  {
    printf("ERROR: Environment variable CCN_GATEWAY_NAME not set\n");
    return 1;
  }
  memset(envGtwName, '\0', sizeof(envGtwName));
  GCN_STRCPY(envGtwName, getenv("CCN_GATEWAY_NAME"));

  if ((returnCode = EVR_OPENLOG(MODULE_PROXY_NAME, MODULE_PROXY_GROUP)) != CSISUCCESS)
  {
    printf("ERROR: EVR_OPENLOG failed returnCode = %ld\n", returnCode);
    return returnCode;
  }

  /*
  Initialize to the OS layer.
  */

  if (CSISUCCESS != OS_init(MODULE_PROXY_NAME,
                            &returnCode,
                            &reasonCode))
  {
    printf("ERROR: OS_init failed returnCode = %ld , reasondCode = %ld\n",returnCode,reasonCode);
  }
  else
  {
    boundOS = CSITRUE;

    /*
    Bind to the CDX layer.
    */

    if (CSISUCCESS != CDX_bind(NULL,
                               NULL,
                               NULL,
                               &returnCode,
                               &reasonCode))
    {
      printf("ERROR: CDX_bind failed returnCode = %ld , reasondCode = %ld\n",returnCode,reasonCode);
    }								
    else
    {
      boundCDX = CSITRUE;
    }
  }

  if (boundOS == CSITRUE && boundCDX == CSITRUE)
  {
    /* Get the CCN WebDomain*/ 

    if (CSISUCCESS != CDX_getWebDomain(NULL, 
                                       &ccnWebDomain, 
                                       &returnCode,
                                       &reasonCode))
    {
      if (verbose)
          printf("ERROR: CDX_getWebDomain failed returnCode = %ld , reasonCode = %ld\n",returnCode,reasonCode);

      EVR_ERRORLOG(EVR_CRITICAL
                   ,(APACHE_MSG_CDX_ERROR
                    ,"CDX_getWebDomain"
                    ,reasonCode
                    )
                   );
      
    }
    else
    {
      strcpy(ccnWebDomain,CCN_DEFAULT_WEB_DOMAIN);
    }

    /* Get Local Gateway Mode */
    if (CSISUCCESS != CDX_getGatewayMode(NULL
                                       ,envSiteName
                                       ,envGtwName
                                       ,&sLocalGtwMode   
                                       ,&returnCode           
                                       ,&reasonCode
                                       )            
       )
    {   
      if (verbose) printf("ERROR: CDX_getGatewayMode failed returnCode %ld , reasonCode %ld\n",returnCode,reasonCode);

      EVR_ERRORLOG(EVR_CRITICAL
                   ,(APACHE_MSG_CDX_ERROR
                    ,"CDX_getGatewayMode"
                    ,reasonCode
                    )
                   );
      return 1;
    }
    /* Get all site */ 

    if (CSISUCCESS != CDX_listSites2(NULL,
                                     &siteList, 
                                     &returnCode,
                                     &reasonCode))
    {
      EVR_ERRORLOG(EVR_CRITICAL
                   ,(APACHE_MSG_CDX_ERROR
                    ,"CDX_listSites2"
                    ,reasonCode
                    )
                   );

      printf("ERROR: CDX_listSites2 failed returnCode = %ld , reasonCode = %ld\n"
            ,returnCode
            ,reasonCode
            );
      CDX_free(NULL, &sLocalGtwMode);
      return 1;
    }

    for (szSiteName = CDX_list_first(NULL, &siteList)
        ;NULL != szSiteName
        ;szSiteName = CDX_list_next(NULL, &siteList)
        )
    {
        CDXSITEATTINFO siteInfo;
        CSICHAR256 szSiteDN;
        CDXRESULT webHostList = {CDXRESULT_DEFAULT};
        CSICHAR *webHostNod;

        if (verbose) { printf("Currently processing site [%s]\n", szSiteName); }

        if (CSISUCCESS != CDX_getAttributesOfSite(NULL
                                                 ,szSiteName
                                                 ,&siteInfo
                                                 ,&returnCode
                                                 ,&reasonCode
                                                 )
           )
        {
            printf("ERROR: CDX_getAttributesOfSite(%s) failed returnCode %ld, reasonCode %ld\n"
                  ,szSiteName
                  ,returnCode
                  ,reasonCode
                  );         

            EVR_ERRORLOG(EVR_CRITICAL
                        ,(APACHE_MSG_CDX_ERROR
                         ,"CDX_getGatewayMode"
                         ,reasonCode
                         )          
                        );
            CDX_free(NULL, &sLocalGtwMode);
            return 1;
        }

        /* Get all Webhost node  */
				/*
        snprintf(szSiteDN
                ,sizeof(szSiteDN)
                ,"nod=%s,%s"
                ,szSiteName
                ,getenv("CCN_ROOT_NAME") == NULL ? "" : getenv("CCN_ROOT_NAME")
                );
				*/
        if (CSISUCCESS != CDX_buildSiteDName(szSiteName
                                            ,&szSiteDN 
                                            ,&returnCode
                                            ,&reasonCode
                                            )
					 )
        {
            sprintf(strError,"SiteName [%s]", szSiteName);

            EVR_ERRORLOG(EVR_CRITICAL
                      ,(APACHE_MSG_CDX_PARAM_ERROR
                       ,"CDX_buildSiteDName"
                       ,reasonCode
                       ,strError  
                       )        
                      );                                                                                                                                                  
        
            printf("ERROR: CDX_buildSiteDName(%s) failed returnCode = %ld , reasonCode = %ld\n"
                  ,szSiteName
                  ,returnCode
                  ,reasonCode
                  );
            CDX_free(NULL, &sLocalGtwMode);
            return 1;
        }

        if (CSISUCCESS != CDX_listWebhostOfSite(NULL
                                               ,szSiteDN
                                               ,&webHostList
                                               ,&returnCode
                                               ,&reasonCode
                                               )
           )
        {
            sprintf(strError,"SiteName [%s]", szSiteName);

            EVR_ERRORLOG(EVR_CRITICAL
                      ,(APACHE_MSG_CDX_PARAM_ERROR
                       ,"CDX_listWebhostOfSite"
                       ,reasonCode
                       ,strError
                       )
                      );

            printf("ERROR: CDX_listWebhostOfSite(%s) failed returnCode = %ld , reasonCode = %ld\n"
                  ,szSiteName
                  ,returnCode
                  ,reasonCode
                  );
            CDX_free(NULL, &sLocalGtwMode);
            return 1;
        }

        for (webHostNod = CDX_list_first(NULL, &webHostList)
            ;NULL != webHostNod
            ;webHostNod = CDX_list_next(NULL, &webHostList)
            )
        {
            CSICHAR48 webHostNod48;
            CSICHAR48 webHostNodCopy;
            CSILONG lenOfPart1;
            CSICHAR48 webHostPart1;
            CSICHAR48 webHostPart2;
            CDXWEBHOSTATTINFO sWebHostAttributes;
            PROXYHEADER proxyInfoHeader;
            CSIBOOLEAN matchingFound=CSIFALSE;

            if (verbose) { printf("\tCurrently processing webhost [%s]\n", webHostNod); }
            memset(webHostNod48, '\0', sizeof(webHostNod48));
            GCN_STRCPY(webHostNod48, webHostNod);
            /* Split the webhost into two fields */
            memset(webHostPart1, '\0', sizeof(webHostPart1));
            memset(webHostPart2, '\0', sizeof(webHostPart2));
            strcpy(webHostNodCopy, webHostNod48);
            strcpy(webHostPart1, strtok(webHostNodCopy, WEBHOST_SEPARATOR));
            strcpy(webHostPart2, strtok(NULL, WEBHOST_SEPARATOR));
            /* If the last character of webHostPart1 is '(', remove it */
            lenOfPart1 = strlen(webHostPart1);
            if (webHostPart1[lenOfPart1-1] == '(')
            {
              webHostPart1[lenOfPart1-1] = '\0';
            }

            /* Get the attributes of the Webhost */
            if (CSISUCCESS != CDX_getAttributesOfWebHost(NULL
                                                        ,szSiteDN
                                                        ,webHostNod48
                                                        ,&sWebHostAttributes
                                                        ,&returnCode
                                                        ,&reasonCode
                                                        )
               )
            {
              sprintf(strError,"SiteName [%s], Webhost [%s]", szSiteDN, webHostNod48);

              EVR_ERRORLOG(EVR_CRITICAL
                      ,(APACHE_MSG_CDX_PARAM_ERROR
                      ,"CDX_getAttributesOfWebHost"
                      ,reasonCode
                      ,strError
                      )
                     );

              printf("ERROR: CDX_getAttributesOfWebHost(%s,%s) failed returnCode = %ld , reasonCode = %ld\n"
                    ,szSiteDN
                    ,webHostNod48
                    ,returnCode
                    ,reasonCode
                    );
              CDX_free(NULL, &sLocalGtwMode);
              CDX_free(NULL, &webHostList);
              return 1;
            }

            if (   sWebHostAttributes.szProxyRemoteName[0] != '\0'
                && (   strcmp(envSiteName, szSiteName)
                    || strcmp(envGtwName, sWebHostAttributes.szSiteName)
                   )
               )    
            {   
                CSILONG x;
                CSIBOOLEAN modeCheck = CSIFALSE;
                CDXRESULT modeList = {CDXRESULT_DEFAULT};
                CSICHAR *szLocalMode, *szMode;
  
                if (CSISUCCESS != CDX_getGatewayMode(NULL
                                                    ,szSiteName
                                                    ,sWebHostAttributes.szSiteName
                                                    ,&modeList
                                                    ,&returnCode
                                                    ,&reasonCode
                                                    )
                   )      
                {  
                    sprintf(strError
                           ,"SiteName [%s], GatewayName [%s]"
                           ,szSiteName
                           ,sWebHostAttributes.szSiteName
                           );

                    EVR_ERRORLOG(EVR_CRITICAL
                                ,(APACHE_MSG_CDX_PARAM_ERROR
                                 ,"CDX_getGatewayMode"
                                 ,reasonCode
                                 ,strError
                                 )
                                );

                    printf("ERROR: CDX_getGatewayMode(%s,%s) failed returnCode = %ld , reasonCode = %ld\n"
                          ,szSiteName
                          ,sWebHostAttributes.szSiteName
                          ,returnCode
                          ,reasonCode
                          );
                    CDX_free(NULL, &sLocalGtwMode);
                    CDX_free(NULL, &webHostList);
                    return 1;
                }

                for (modeCheck = CSIFALSE, szLocalMode = CDX_list_first(NULL, &sLocalGtwMode)
                    ;CSIFALSE == modeCheck && NULL != szLocalMode
                    ;szLocalMode = CDX_list_next(NULL, &sLocalGtwMode)
                    )
                {
                    for (szMode = CDX_list_first(NULL, &modeList)
                        ;CSIFALSE == modeCheck && NULL != szMode
                        ;szMode = CDX_list_next(NULL, &modeList)
                        )
                    {
                        if (!strcasecmp(szMode, szLocalMode))
                            modeCheck = CSITRUE;
                    }
                }
                CDX_free(NULL, &modeList);
                if (CSIFALSE == modeCheck)
                    continue;
                
                if (verbose)
                {           
                    printf("WebhostNod = %s, webHostPart1 = %s, webHostPart1 = %s\n"
                          ,webHostNod48
                          ,webHostPart1
                          ,webHostPart1
                          );           
                    printf("ccnSiteName = %s, ccnHostName = %s, ccnProxyRemote = %s, ccnIsHttps = %s\n"
                          ,sWebHostAttributes.szSiteName
                          ,sWebHostAttributes.szHostName
                          ,sWebHostAttributes.szProxyRemoteName
                          ,sWebHostAttributes.isHttps ? "1" : "0"
                          );
                }         
                strcpy(proxyInfoHeader.szWebHostPart1,webHostPart1);
                strcpy(proxyInfoHeader.szWebSite,siteInfo.szWebSiteName);
                strcpy(proxyInfoHeader.szWebDomain,ccnWebDomain);
                strcpy(proxyInfoHeader.szProxyRemote,sWebHostAttributes.szProxyRemoteName);
                proxyInfoHeader.isHttps = sWebHostAttributes.isHttps;
  
                for (x = 0 ; x < NB_PATTERN_PROXY ; x++)
                {    
                    if (!validate_pattern(&proxy_remote_build[x],
                                          webHostPart2,          
                                          &proxyInfoHeader))
                    {  
                      matchingFound = CSITRUE;
                      break;
                    }       
                }  
                if (matchingFound == CSIFALSE)
                {    
                    EVR_ERRORLOG(EVR_CRITICAL
                      ,(APACHE_WRONG_REGULAR_EXPR
                       ,webHostNod48
                       )            
                      );
                    printf("ERROR: Invalid Regular expression used in the Webhost entry (%s)\n",
                                            webHostNod48);
                }    
            }      
        }
        CDX_free(NULL, &webHostList);
    }
    CDX_free(NULL, &siteList);
    CDX_free(NULL,&sLocalGtwMode);
  }

  /*
  Unbind from the CDX layer.
  */

  if (boundCDX == CSITRUE)
  {
    CDX_unbind(NULL,
               &returnCode,
               &reasonCode);
  }

  /*
  Release to the OS layer.
  */

  if (boundOS == CSITRUE)
  {
    OS_term(&returnCode,
            &reasonCode);
  }

  EVR_CLOSELOG();
  return 0;
}	

/*
* @EXITVALUE CSILONG
* gen_proxy_remote exits 0 upon success
*/

int main(int argc,char *argv[])
{
   int exitStatus = 0;
   int verbose = 0;

   if (argc > 1 && ((char)(argv[1][0]) != (char)'-'))
   {
      usage ();
   }
   else
   {
      if (argc > 1)
      {
         switch ((int)(argv[1][1]))
         {
            case 'v':
               verbose = 1;
               break;

            default:
               verbose = 0;
         }
      }

      exitStatus = gen_proxy_remote(verbose);
   }

   return (exitStatus);
}
